# MIT License
#
# Copyright (c) 2023 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import TermTk as ttk

from .superobj.superwidget import SuperWidget
from .superobj.superlayout import SuperLayout

class _SignalSlotItem(ttk.TTkTreeWidgetItem):
    __slots__ = ('_sender', '_signal', '_receiver', '_slot', '_designer', '_signalData', '_slotData', '_avoidRecursion')
    def __init__(self, designer, *args, **kwargs):
        self._designer = designer
        self._sender   = ttk.TTkComboBox(height=1, list=['<sender>'  ], index=0)
        self._signal   = ttk.TTkComboBox(height=1, list=['<signal>'  ], index=0)
        self._receiver = ttk.TTkComboBox(height=1, list=['<receiver>'], index=0)
        self._slot     = ttk.TTkComboBox(height=1, list=['<slot>'    ], index=0)
        self._signalData = {}
        self._slotData = {}
        self._avoidRecursion = False
        self._sender.currentTextChanged.connect(  self._senderChanged)
        self._signal.currentTextChanged.connect(  self._signalChanged)
        self._receiver.currentTextChanged.connect(self._receiverChanged)
        self._slot.currentTextChanged.connect(    self._slotChanged)
        self._designer.widgetNameChanged.connect( self._widgetNameChanged)
        self._designer.weModified.connect(self.updateWidgets)
        self.updateWidgets()
        super().__init__([self._sender,self._signal,self._receiver,self._slot], *args, **kwargs)

    def isValid(self):
        curSender   = str(self._sender.currentText())
        curReceiver = str(self._receiver.currentText())
        curSignal   = str(self._signal.currentText())
        curSlot     = str(self._slot.currentText())
        if curSender=='<sender>' or curReceiver == '<receiver>':
            return False
        ret = False
        for ccName in self._signalData:
            if curSignal in self._signalData[ccName]:
                ret = True
                break
        for ccName in self._slotData:
            if curSlot in self._slotData[ccName]:
                ret &= True
                break
        return ret

    def importConnection(self, connection):
        self._sender.setCurrentText(connection['sender'])
        self._receiver.setCurrentText(connection['receiver'])
        self._signal.setCurrentText(connection['signal'])
        self._slot.setCurrentText(connection['slot'])

    def dumpDict(self):
        curSender   = str(self._sender.currentText())
        curReceiver = str(self._receiver.currentText())
        curSignal   = str(self._signal.currentText())
        curSlot     = str(self._slot.currentText())
        return {
            'sender':   curSender,
            'receiver': curReceiver,
            'signal':   curSignal,
            'slot':     curSlot }

    @ttk.pyTTkSlot()
    def updateWidgets(self):
        self._widgetNameChanged()

    @staticmethod
    def typeToString(t):
        if type(t) in (list,tuple):
            return ",".join([_SignalSlotItem.typeToString(x) for x in t])
        return {bool    : 'bool',
                int     : 'int',
                float   : 'float',
                complex : 'complex',
                str     : 'str',
                None    : ''}.get(t,f"UNDEFINED {t}")

    @ttk.pyTTkSlot(str, str)
    def _widgetNameChanged(self, oldName='', newName=''):
        names = [w.name() for w in self._designer.getWidgets()]

        def _setCB(cb:ttk.TTkComboBox, base):
            text = cb.currentText()
            text = text if text!=oldName else newName
            cb.clear()
            cb.addItems([base]+names)
            cb.setCurrentText(text)

        _setCB(self._sender,   '<sender>')
        _setCB(self._receiver, '<receiver>')

    @ttk.pyTTkSlot(str)
    def _senderChanged(self, text):
        self._signalData, _ = self.getSignalSlot(text)
        self._refreshSignals()

    def _refreshSignals(self):
        curSignal = str(self._signal.currentText())
        curSlot = str(self._slot.currentText())
        filter = None
        for c in self._slotData:
            if not str(curSlot) in self._slotData[c]: continue
            filter = self._slotData[c][curSlot]['type']
            break

        signals = ['<signal>']
        for ccName in self._signalData:
            signals.append(ttk.TTkString(f"{ccName}:",ttk.TTkColor.fg("#FFFF88")+ttk.TTkColor.UNDERLINE))
            for s in self._signalData[ccName]:
                if filter in (None,self._signalData[ccName][s]['type']):
                    signals.append(s)
        self._signal.clear()
        self._signal.addItems(signals)
        self._signal.setCurrentText(curSignal)

    @ttk.pyTTkSlot(str)
    def _signalChanged(self, text):
        if self._avoidRecursion: return
        self._avoidRecursion = True
        self._refreshSlots()
        self._avoidRecursion = False


    @ttk.pyTTkSlot(str)
    def _receiverChanged(self, text):
        _, self._slotData = self.getSignalSlot(text)
        self._refreshSlots()

    def _refreshSlots(self):
        curSignal = self._signal.currentText()
        curSlot = self._slot.currentText()
        filter = 'ALL'
        for c in self._signalData:
            if not str(curSignal) in self._signalData[c]: continue
            filter = self._signalData[c][curSignal]['type']
            break

        slots = ['<slot>']
        for ccName in self._slotData:
            slots.append(ttk.TTkString(f"{ccName}:",ttk.TTkColor.fg("#FFFF88")+ttk.TTkColor.UNDERLINE))
            for s in self._slotData[ccName]:
                if filter in ('ALL',self._slotData[ccName][s]['type']) or not self._slotData[ccName][s]['type']:
                    slots.append(s)

        self._slot.clear()
        self._slot.addItems(slots)
        self._slot.setCurrentText(curSlot)

    @ttk.pyTTkSlot(str)
    def _slotChanged(self, text):
        if self._avoidRecursion: return
        self._avoidRecursion = True
        self._refreshSignals()
        self._avoidRecursion = False

    def getSignalSlot(self, name):
        if not (widget := {w.name():w for w in self._designer.getWidgets()}.get(name,None)):
            return {},{}
        ttk.TTkLog.debug(f"Selected {widget=}")
        slots = {}
        signals = {}
        for cc in reversed(type(widget).__mro__):
            if issubclass(cc, ttk.TTkWidget) or issubclass(cc, ttk.TTkLayout):
                ccName = cc.__name__
                if ccName in ttk.TTkUiProperties:
                    if ttk.TTkUiProperties[ccName]['slots']:
                        slots[ccName] = ttk.TTkUiProperties[ccName]['slots']
                    if ttk.TTkUiProperties[ccName]['signals']:
                        signals[ccName] = ttk.TTkUiProperties[ccName]['signals']
        return signals,slots

class SignalSlotEditor(ttk.TTkWidget):
    __slots__ = ('_items', '_designer')
    def __init__(self, designer, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setLayout(ttk.TTkGridLayout())
        self._items = []
        self._designer = designer

        self._detail = ttk.TTkTree()
        self._detail.setHeaderLabels(["Sender","Signal","Receiver","Slot"])
        self.layout().addWidget(addb := ttk.TTkButton(border=False, text="ADD"),1,1)
        self.layout().addWidget(delb := ttk.TTkButton(border=False, text="DEL"),1,2)
        self.layout().addWidget(ttk.TTkLabel(text=" Signal/Slot Editor "),1,3)
        self.layout().addWidget(self._detail,2,1,1,3)

        addb.clicked.connect(self._addStuff)
        delb.clicked.connect(self._delStuff)

    def _addStuff(self):
        item = _SignalSlotItem(self._designer)
        self._items.append(item)
        self._detail.addTopLevelItem(item)

    def _delStuff(self):
        if not (items := self._detail.selectedItems()):
            return
        for item in items:
            if (index := self._detail.indexOfTopLevelItem(item)) is not None:
                self._detail.takeTopLevelItem(index)

    def importConnections(self, connections):
        self._items = []
        self._detail.clear()
        for c in connections:
            item = _SignalSlotItem(self._designer)
            item.importConnection(c)
            self._items.append(item)
            self._detail.addTopLevelItem(item)

    def dumpDict(self):
        ret = []
        for i in self._items:
            if i.isValid():
                ret.append(i.dumpDict())
        return ret



# MIT License
#
# Copyright (c) 2023 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import TermTk as ttk

from .superobj.superwidget import SuperWidget
from .superobj.superlayout import SuperLayout

class _TTkTomTreeWidgetItem(ttk.TTkTreeWidgetItem):
    __slots__ = ('_tomWidget','_tomSuperWidget')
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tomWidget = kwargs.get('tomWidget')
        self._tomSuperWidget = kwargs.get('tomSuperWidget')
    def tomWidget(self):
        return self._tomWidget
    def tomSuperWidget(self):
        return self._tomSuperWidget

class TreeInspector(ttk.TTkGridLayout):
    __slots__ = ('_windowEditor','_tomTree', '_designer')
    def __init__(self, designer, windowEditor, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._designer = designer
        self._windowEditor = windowEditor

        self._tomTree = ttk.TTkTree()
        self._tomTree.setHeaderLabels(["Object", "Class", "Visibility", "Layout"])
        self._tomTree.addTopLevelItem(TreeInspector._getTomTreeItem(windowEditor.getTTk().widgetItem()))

        @ttk.pyTTkSlot(_TTkTomTreeWidgetItem, int)
        def _itemSelected(wi, _):
            if tomw :=  wi.tomWidget():
                self._designer.thingSelected.emit(tomw, wi.tomSuperWidget())

        self._tomTree.itemClicked.connect(_itemSelected)

        btnRefresh = ttk.TTkButton(text="Refresh")
        btnRefresh.clicked.connect(self.refresh)

        btnPick = ttk.TTkButton(text="Pick 🔍")
        # btnPick.clicked.connect(self._btnPickCb)

        self.addWidget(self._tomTree, 1,0, 1,3)
        self.addWidget(btnPick,       0,0, 1,1)
        self.addWidget(btnRefresh,    0,1, 1,2)

    #@ttk.pyTTkSlot(ttk.TTkKeyEvent, ttk.TTkMouseEvent)
    #def _processInput(self, kevt, mevt):
    #    ttk.TTkLog.debug(f"{kevt} {mevt}")
    #    if mevt.evt == TTkK.Press:
    #        # TTkHelper._rootWidget.setEnabled(True)
    #        # TTkHelper._rootWidget._input.inputEvent.connect(TTkHelper._rootWidget._processInput)
    #        ttk.TTkHelper._rootWidget._input.inputEvent.disconnect(self._processInput)
    #        thing = TTkTomInspector._findWidget(mevt,ttk.TTkHelper._rootWidget.rootLayout())
    #        ttk.TTkLog.debug(f"{thing=}")
    #        if thing:
    #            self._makeDetail(thing)
    #        else:
    #            self._detail = ttk.TTkFrame(title=f"None")
    #        self._splitter.replaceWidget(1,self._detail)
    #        self._refresh(thing)

    @ttk.pyTTkSlot()
    def _btnPickCb(self):
        # ttk.TTkHelper._rootWidget._input.inputEvent.connect(self._processInput)
        pass

    @ttk.pyTTkSlot()
    def refresh(self, thing=None):
        self._tomTree.clear()
        self._tomTree.addTopLevelItem(TreeInspector._getTomTreeItem(self._windowEditor.getTTk().widgetItem()))

    @staticmethod
    def _getTomTreeItem(layoutItem, widSelected=None):
        if layoutItem.layoutItemType == ttk.TTkK.WidgetItem:
            superThing = thing = layoutItem.widget()
            if issubclass(type(superThing), SuperWidget):
                thing = thing._wid
            elif issubclass(type(superThing), SuperLayout):
                thing = thing._lay
            expanded = True # ttk.TTkHelper.isParent(widSelected,thing) if widSelected else False
            if issubclass(type(superThing), SuperWidget):
                top = _TTkTomTreeWidgetItem([
                            thing._name,   thing.__class__.__name__,
                            str(thing.isVisible()),
                            thing.layout().__class__.__name__],
                            tomWidget=thing,
                            tomSuperWidget=superThing,
                            expanded=expanded)
            elif issubclass(type(superThing), SuperLayout):
                top = _TTkTomTreeWidgetItem([
                            'Layout',   thing.__class__.__name__,
                            '',
                            thing.__class__.__name__],
                            tomWidget=thing,
                            tomSuperWidget=superThing,
                            expanded=expanded)
            if issubclass(type(superThing), SuperWidget):
                for c in superThing._superLayout.layout().children():
                    top.addChild(TreeInspector._getTomTreeItem(c,widSelected))
            elif issubclass(type(superThing), SuperLayout):
                for c in superThing.layout().children():
                    top.addChild(TreeInspector._getTomTreeItem(c,widSelected))
            else:
                for c in superThing.layout().children():
                    top.addChild(TreeInspector._getTomTreeItem(c,widSelected))

            # for c in thing.rootLayout().children():
            #     if c == thing.layout(): continue
            #     if c.layoutItemType == ttk.TTkK.LayoutItem:
            #         top.addChild(tc:=_TTkTomTreeWidgetItem(["layout (Other)", c.__class__.__name__, ""]))
            #         for cc in c.children():
            #             tc.addChild(TreeInspector._getTomTreeItem(cc,widSelected))
            return top

        if layoutItem.layoutItemType == ttk.TTkK.LayoutItem:
            top = _TTkTomTreeWidgetItem(["layout", layoutItem.__class__.__name__,"",layoutItem.__class__.__name__])
            for c in layoutItem.children():
                top.addChild(TreeInspector._getTomTreeItem(c,widSelected))
            return top

        return _TTkTomTreeWidgetItem(["ERROR!!!", "None"])


# MIT License
#
# Copyright (c) 2023 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import json

from TermTk import TTk, TTkK, TTkLog, TTkCfg, TTkColor, TTkTheme, TTkTerm, TTkHelper
from TermTk import TTkString
from TermTk import TTkColorGradient
from TermTk import pyTTkSlot, pyTTkSignal

from TermTk import TTkWidget, TTkFrame, TTkButton, TTkLabel
from TermTk import TTkTabWidget
from TermTk import TTkAbstractScrollArea, TTkAbstractScrollView, TTkScrollArea
from TermTk import TTkFileDialogPicker, TTkMessageBox
from TermTk import TTkFileTree, TTkTextEdit

from TermTk import TTkLayout, TTkGridLayout, TTkVBoxLayout, TTkHBoxLayout
from TermTk import TTkSplitter
from TermTk import TTkLogViewer, TTkTomInspector

from TermTk import TTkUiLoader, TTkUtil

from .cfg  import *
from .about import *
from .widgetbox import DragDesignItem, WidgetBox, WidgetBoxScrollArea
from .windoweditor import WindowEditor, SuperWidget
from .treeinspector import TreeInspector
from .propertyeditor import PropertyEditor
from .signalsloteditor import SignalSlotEditor
from .quickexport import QuickExport

#
#      Mimic the QT Designer layout
#
#      ┌─────────────────────╥───────────────────────────────╥───────────────────┐
#      │                     ║┌─────────────────────────────┐║                   │
#      │                     ║│       ToolBar               │║  Tree Inspector   │
#      │                     ║├─────────────────────────────┤║                   │
#      │                     ║│                             │║                   │
#      │   Widget            ║│                             │║                   │
#      │   Box               ║│                             │╠═══════════════════╡
#      │                     ║│      Main Window            │║                   │
#      │                     ║│      Editor                 │║   Property        │
#      │                     ║│                             │║   Editor          │
#      │                     ║│                             │║                   │
#      │                     ║│                             │║                   │
#      │                     ║│                             │║                   │
#      │                     ║│                             │║                   │
#      │                     ║│                             │╠═══════════════════╡
#      │                     ║│                             │║                   │
#      │                     ║└─────────────────────────────┘║   Signal/Slot     │
#      │                     ╟───────────────────────────────╢                   │
#      │                     ║     LOG Viewer                ║   Editor          │
#      │                     ║                               ║                   │
#      └─────────────────────╨───────────────────────────────╨───────────────────┘
#

class TTkDesigner(TTkGridLayout):
    __slots__ = ('_pippo', '_main', '_windowEditor', '_toolBar', '_fileNameLabel', '_sigslotEditor', '_treeInspector', '_fileName', '_currentPath',
                 # Signals
                 'weModified', 'thingSelected', 'widgetNameChanged'
                 )
    def __init__(self, *args, **kwargs):
        self._fileName = "untitled.tui.json"
        self._currentPath = self._currentPath =  os.path.abspath('.')

        self.weModified = pyTTkSignal()
        self.thingSelected = pyTTkSignal(TTkWidget, TTkWidget)
        self.widgetNameChanged = pyTTkSignal(str, str)

        super().__init__(*args, **kwargs)

        self.addWidget(mainSplit := TTkSplitter())
        mainSplit.addItem(widgetBoxLayout := TTkVBoxLayout())
        #mainSplit.addWidget(TTkButton(text='A',border=True))

        # mainSplit.addWidget(sa := TTkScrollArea())
        # sa.viewport().setLayout(TTkGridLayout())
        # sa.viewport().layout().addWidget(WindowEditor())

        self._main = TTkVBoxLayout()
        self._toolBar = TTkHBoxLayout()
        self._windowEditor = WindowEditor(self)
        self._sigslotEditor = SignalSlotEditor(self)
        self._treeInspector = TreeInspector(self, self._windowEditor.viewport())
        self._fileNameLabel = TTkLabel(text=f"( {self._fileName} )", alignment=TTkK.CENTER_ALIGN)

        widgetBoxLayout.addWidget(topMenuFrame := TTkFrame(minHeight=1,maxHeight=1,border=False))
        widgetBoxLayout.addWidget(WidgetBoxScrollArea(self))

        self._main.addItem(self._toolBar)
        self._main.addWidget(self._windowEditor)

        mainSplit.addWidget(centralSplit := TTkSplitter(orientation=TTkK.VERTICAL))
        centralSplit.addWidget(self._main)
        centralSplit.addWidget(bottonTabWidget := TTkTabWidget(border=False))
        # centralSplit.addWidget(TTkLogViewer())
        bottonTabWidget.addTab(self._sigslotEditor,'Signal/Slot Editor')
        bottonTabWidget.addTab(TTkLogViewer(),'Logs')

        mainSplit.addWidget(rightSplit := TTkSplitter(orientation=TTkK.VERTICAL))

        rightSplit.addItem(self._treeInspector)
        rightSplit.addItem(propertyEditor := PropertyEditor())
        # rightSplit.addItem(self._sigslotEditor)

        self.thingSelected.connect(lambda _,s : s.pushSuperControlWidget())
        self.thingSelected.connect(propertyEditor.setDetail)

        self.weModified.connect(self._treeInspector.refresh)

        fileMenu = topMenuFrame.menubarTop().addMenu("&File")
        fileMenu.addMenu("New").menuButtonClicked.connect(self.new)
        fileMenu.addMenu("Open").menuButtonClicked.connect(self.open)
        fileMenu.addMenu("Save").menuButtonClicked.connect(self.save)
        fileMenu.addMenu("Save As...").menuButtonClicked.connect(self.saveAs)
        fileMenu.addMenu("Exit").menuButtonClicked.connect(TTkHelper.quit)

        fileMenu = topMenuFrame.menubarTop().addMenu("F&orm")
        fileMenu.addMenu("Preview...").menuButtonClicked.connect(self.preview)

        def _showAbout(btn):
            TTkHelper.overlay(None, About(), 30,10)
        def _showAboutTTk(btn):
            TTkHelper.overlay(None, TTkAbout(), 30,10)

        helpMenu = topMenuFrame.menubarTop().addMenu("&Help", alignment=TTkK.RIGHT_ALIGN)
        helpMenu.addMenu("About ...").menuButtonClicked.connect(_showAbout)
        helpMenu.addMenu("About ttk").menuButtonClicked.connect(_showAboutTTk)

        w,_ = self.size()
        mainSplit.setSizes([25,None,40])
        centralSplit.setSizes([None,8])

        self._toolBar.addWidget(btnPreview := TTkButton(maxWidth=12, text='Preview...'))
        self._toolBar.addWidget(btnExport := TTkButton(maxWidth=17, text='Quick Export 📦'))
        self._toolBar.addWidget(btnColors  := TTkButton(maxWidth=11, checkable=True, text=
                            TTkString("▣",TTkColor.fg("#ff0000")) +
                            TTkString("▣",TTkColor.fg("#ffff00")) +
                            TTkString("▣",TTkColor.fg("#00ff00")) +
                            TTkString("▣",TTkColor.fg("#00ffff")) +
                            TTkString("▣",TTkColor.fg("#0000ff")) + "🦄"))

        btnPreview.clicked.connect(self.preview)
        btnExport.clicked.connect(self.quickExport)
        btnColors.toggled.connect(self.toggleColors)

        self._toolBar.addWidget(self._fileNameLabel)


    def getWidgets(self):
        widgets = []
        def _getItems(layoutItem):
            if layoutItem.layoutItemType == TTkK.WidgetItem:
                superThing = layoutItem.widget()
                if issubclass(type(superThing), SuperWidget):
                    widgets.append(superThing._wid)
                for c in superThing.layout().children():
                    _getItems(c)
        _getItems(self._windowEditor.getTTk().widgetItem())
        return widgets

    @pyTTkSlot(bool)
    def toggleColors(self, state):
        SuperWidget.toggleHighlightLayout.emit(state)

    @pyTTkSlot()
    def quickExport(self):
        tui = self._windowEditor.dumpDict()
        connections = self._sigslotEditor.dumpDict()
        data = {
            'version':'1.0.0',
            'tui':tui,
            'connections':connections}

        win = QuickExport(
                data=data,
                title="Mr Export", size=(80,30),
                flags=TTkK.WindowFlag.WindowMaximizeButtonHint|TTkK.WindowFlag.WindowCloseButtonHint)
        TTkHelper.overlay(None, win, 2, 2, modal=True)

    @pyTTkSlot()
    def preview(self):
        tui = self._windowEditor.dumpDict()
        connections = self._sigslotEditor.dumpDict()
        # for line in jj.split('\n'):
        #     TTkLog.debug(f"{line}")
        newUI = {
            'version':'1.0.0',
            'tui':tui,
            'connections':connections}
        jj =  json.dumps(newUI, indent=1)

        widget = TTkUiLoader.loadJson(jj)
        win = TTkWindow(
                title="Mr Terminal",
                size=(80,30),
                layout=TTkGridLayout(),
                flags=TTkK.WindowFlag.WindowMaximizeButtonHint|TTkK.WindowFlag.WindowCloseButtonHint)
        win.layout().addWidget(widget)
        TTkHelper.overlay(None, win, 2, 2, modal=True)

    @pyTTkSlot()
    def new(self):
        newWindow = TTkUiLoader.loadFile(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../tui/newWindow.tui.json"))
        newWindow.getWidgetByName("BtnWindow").clicked.connect(self._windowEditor.newWindow)
        newWindow.getWidgetByName("BtnWidget").clicked.connect(self._windowEditor.newWidget)
        newWindow.getWidgetByName("BtnWindow").clicked.connect(self.weModified.emit)
        newWindow.getWidgetByName("BtnWidget").clicked.connect(self.weModified.emit)
        TTkHelper.overlay(self._windowEditor, newWindow, 10, 4, modal=True)

    def _openFile(self, fileName):
        TTkLog.info(f"Open: {fileName}")
        self._currentPath =  os.path.dirname(os.path.abspath(fileName))
        self._fileName =  os.path.basename(os.path.abspath(fileName))
        self._fileNameLabel.setText(f"( {self._fileName} )")

        with open(fileName) as fp:
            dd = json.load(fp)
            sw = SuperWidget.loadDict(self, self._windowEditor.viewport(), dd['tui'])
            self._windowEditor.importSuperWidget(sw)
            self._sigslotEditor.importConnections(dd['connections'])
            self._treeInspector.refresh()

    @pyTTkSlot()
    def open(self):
        filePicker = TTkFileDialogPicker(pos = (3,3), size=(80,30), caption="Open", path=self._currentPath, fileMode=TTkK.FileMode.ExistingFile ,filter="TTk Tui Files (*.tui.json);;Json Files (*.json);;All Files (*)")
        filePicker.pathPicked.connect(self._openFile)
        TTkHelper.overlay(None, filePicker, 5, 5, True)

    @pyTTkSlot()
    def save(self):
        return self._saveToFile(os.path.join(self._currentPath,self._fileName))

    def _saveToFile(self, fileName):
        TTkLog.info(f"Saving to: {fileName}")

        tui = self._windowEditor.dumpDict()
        connections = self._sigslotEditor.dumpDict()
        newUI = {
            'version':'1.0.0',
            'tui':tui,
            'connections':connections}
        jj =  json.dumps(newUI, indent=1)

        with open(fileName,'w') as fp:
            fp.write(jj)

    @pyTTkSlot()
    def saveAs(self):
        def _approveFile(fileName):
            if os.path.exists(fileName):
                messageBox = TTkMessageBox(
                    title='Title',
                    text= (
                        TTkString( f'A file named "{os.path.basename(fileName)}" already exists.\nDo you want to replace it?', TTkColor.BOLD) +
                        TTkString( f'\n\nReplacing it will overwrite its contents.') ),
                    icon=TTkMessageBox.Icon.Warning,
                    standardButtons=TTkMessageBox.StandardButton.Discard|TTkMessageBox.StandardButton.Save|TTkMessageBox.StandardButton.Cancel)
                messageBox.buttonSelected.connect(lambda btn : self._saveToFile(fileName) if btn == TTkMessageBox.StandardButton.Save else None)
                TTkHelper.overlay(None, messageBox, 5, 5, True)
            else:
                self._saveToFile(fileName)
        filePicker = TTkFileDialogPicker(pos = (3,3), size=(80,30), caption="Save As...", path=os.path.join(self._currentPath,self._fileName), fileMode=TTkK.FileMode.AnyFile ,filter="TTk Tui Files (*.tui.json);;Json Files (*.json);;All Files (*)")
        filePicker.pathPicked.connect(_approveFile)
        TTkHelper.overlay(None, filePicker, 5, 5, True)



# MIT License
#
# Copyright (c) 2023 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .superobj import SuperWidget

import TermTk as ttk


class WindowEditorView(ttk.TTkAbstractScrollView):
    __slots__ = ('_designer', '_snapRootWidget', '_ttk')
    def __init__(self, designer, *args, **kwargs):
        self._designer = designer
        super().__init__(*args, **kwargs)
        self._ttk = None
        self.viewChanged.connect(self._viewChangedHandler)
        self.newWidget()

    @ttk.pyTTkSlot()
    def newWindow(self):
        if self._ttk:
            self._ttk.superResized.disconnect(self._superChanged)
            self._ttk.superMoved.disconnect(self._superChanged)
            self.layout().removeWidget(self._ttk)
        self._ttk = SuperWidget(wid=ttk.TTkWindow(name = 'MainWindow'), designer=self._designer, pos=(4,2), superRootWidget=True)
        self._ttk.resize(self.width()-8,self.height()-4)
        self._snapRootWidget = True
        self.layout().addWidget(self._ttk)
        self._ttk.superResized.connect(self._superChanged)
        self._ttk.superMoved.connect(self._superChanged)

    @ttk.pyTTkSlot()
    def newWidget(self):
        if self._ttk:
            self._ttk.superResized.disconnect(self._superChanged)
            self._ttk.superMoved.disconnect(self._superChanged)
            self.layout().removeWidget(self._ttk)
        self._ttk = SuperWidget(wid=ttk.TTkWidget(name = 'MainWidget'), designer=self._designer, pos=(4,2), superRootWidget=True)
        self._ttk.resize(self.width()-8,self.height()-4)
        self._snapRootWidget = True
        self.layout().addWidget(self._ttk)
        self._ttk.superResized.connect(self._superChanged)
        self._ttk.superMoved.connect(self._superChanged)

    def importWidget(self, wid):
        if self._ttk:
            self._ttk.superResized.disconnect(self._superChanged)
            self._ttk.superMoved.disconnect(self._superChanged)
        self._snapRootWidget = False
        self._ttk = SuperWidget.swFromWidget(wid=wid, designer=self._designer, pos=(4,2), size=(self.width()-8,self.height()-4))
        self._ttk.makeRootWidget()
        self.layout().addWidget(self._ttk)

    def importSuperWidget(self, sw):
        if self._ttk:
            self.layout().removeWidget(self._ttk)
        self._ttk = sw
        self._ttk.makeRootWidget()
        self.layout().addWidget(self._ttk)

    def getTTk(self):
        return self._ttk

    def dumpDict(self):
        return self._ttk.dumpDict()

    @ttk.pyTTkSlot()
    def _superChanged(self):
        self._snapRootWidget = False
        self.viewChanged.emit()

    def resizeEvent(self, w, h):
        if self._snapRootWidget:
            self._ttk.resize(w-8,h-4)
            self._snapRootWidget = True
        return super().resizeEvent(w, h)

    @ttk.pyTTkSlot()
    def _viewChangedHandler(self):
        x,y = self.getViewOffsets()
        self.layout().setOffset(-x,-y)

    def viewFullAreaSize(self):
        x,y,w,h = self.layout().fullWidgetAreaGeometry()
        return x+w, y+h

    def viewDisplayedSize(self):
        return self.size()

    def paintEvent(self):
        w,h = self.size()
        # self._canvas.fill(pos=(0,0),size=(w,h), char="╳", color=ttk.TTkColor.fg("#444400")+ttk.TTkColor.bg("#000044"))
        self._canvas.fill(pos=(0,0),size=(w,h), char="#", color=ttk.TTkColor.fg("#220044")+ttk.TTkColor.bg("#000022"))

class WindowEditor(ttk.TTkAbstractScrollArea):
    __slots__ = ('getTTk', 'dumpDict', 'importWidget', 'importSuperWidget',
                 # Forwarded slots
                 'newWindow', 'newWidget')
    def __init__(self, designer, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setViewport(wev := WindowEditorView(designer))
        # wev.importWidget(ttk.TTkWidget(name = 'TTk'))
        # Forward Methods
        self.getTTk            = wev.getTTk
        self.dumpDict          = wev.dumpDict
        self.importWidget      = wev.importWidget
        self.importSuperWidget = wev.importSuperWidget
        self.newWindow         = wev.newWindow
        self.newWidget         = wev.newWidget


# MIT License
#
# Copyright (c) 2023 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import TermTk as ttk

class QuickExport(ttk.TTkWindow):
    __slots__ = ('_data', '_te')
    def __init__(self, data, *args, **kwargs):
        self._data = data
        super().__init__(*args, **kwargs|{'layout':ttk.TTkGridLayout()})
        self.layout().addWidget(btnCompressed := ttk.TTkButton(text='Compressed', border=True, maxHeight=3), 0,1)
        self.layout().addWidget(btnNormal     := ttk.TTkButton(text='Normal', border=True, maxHeight=3), 0,0)
        self._te = ttk.TTkTextEdit(lineNumber=True, readOnly=False)
        self.layout().addWidget(self._te, 1,0,1,2)

        btnNormal.clicked.connect(self.useNormmal)
        btnCompressed.clicked.connect(self.useCompressed)

        self.useCompressed()

    ttk.pyTTkSlot()
    def useNormmal(self):
        self._te.setLineWrapMode(ttk.TTkK.WidgetWidth)
        self._te.setWordWrapMode(ttk.TTkK.WordWrap)
        self._te.setText('from TermTk import TTkUtil, TTkUiLoader, TTk\n')
        self._te.append(f'# Data generated using ttkDesigner')
        self._te.append('widget = TTkUiLoader.loadDict(')
        self._te.append(str(self._data)+ '")')
        self._te.append('\nroot=TTk()\nroot.layout().addWidget(widget)\nroot.mainloop()\n')

    ttk.pyTTkSlot()
    def useCompressed(self):
        self._te.setLineWrapMode(ttk.TTkK.NoWrap)
        b64str = ttk.TTkUtil.obj_inflate_2_base64(self._data)
        self._te.setText('from TermTk import TTkUtil, TTkUiLoader, TTk\n')
        self._te.append(f'# Data generated using ttkDesigner')
        self._te.append('widget = TTkUiLoader.loadDict(TTkUtil.base64_deflate_2_obj(')
        b64list = '    "' + '" +\n    "'.join([b64str[i:i+128] for i in range(0,len(b64str),128)]) + '"))'
        self._te.append(b64list)
        self._te.append('\nroot=TTk()\nroot.layout().addWidget(widget)\nroot.mainloop()\n')




# MIT License
#
# Copyright (c) 2023 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import TermTk as ttk

class PropertyEditor(ttk.TTkGridLayout):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._detail = ttk.TTkTree()
        self._detail.setHeaderLabels(["Property","Value"])
        self._widget = ttk.TTkWidget()
        self._superWidget = ttk.TTkWidget()
        self.addWidget(self._detail)

    @ttk.pyTTkSlot(ttk.TTkWidget, ttk.TTkWidget)
    def setDetail(self, widget, superWidget):
        self._widget = widget
        self._superWidget = superWidget
        # TBD
        # Override: 'name', 'radiogroup',
        # ro geometry if grid layout

        self._makeDetail(widget, *superWidget.getSuperProperties())

    def _makeDetail(self, domw, exceptions, exclude):
        def _boundValue(_f,_w,_v):
            def _ret():
                _f(_w,_v)
                self._superWidget.updateAll()
            return _ret
        def _bound(_f,_w,_l):
            def _ret(_v):
                _f(_w,_l(_v))
                self._superWidget.updateAll()
            return _ret
        def _boundFlags(_f,_g,_w,_l,_flag):
            def _ret(_v):
                _val = _g(_w)
                _val = _val|_flag if _l(_v) else _val&~_flag
                _f(_w,_val)
                self._superWidget.updateAll()
            return _ret
        def _boundTextEdit(_f,_w,_te):
            def _ret():
                _v = _te.getTTkString()
                _f(_w,_v)
                self._superWidget.updateAll()
            return _ret
        def _boundLineEdit(_f,_w,_te):
            def _ret(t):
                _v = str(t)
                _f(_w,_v)
                self._superWidget.updateAll()
            return _ret

        # Multi Flag Fields
        # ▼ Input Type     │ - (0x0001)
        #   • Text         │[X] 0x0001
        #   • Number       │[ ] 0x0002
        #   • Password     │[ ] 0x0004
        def _processMultiFlag(name, prop):
            flags = prop['get']['flags']
            ret = ttk.TTkTreeWidgetItem([name,f" - (0x{prop['get']['cb'](domw):04X})"],expanded=True)
            # value = ttk.TTkFrame(layout=ttk.TTkVBoxLayout(), height=len(flags), border=False)
            for fl in flags:
                if 'set' in prop:
                    fcb = ttk.TTkCheckbox(text=f" 0x{flags[fl]:04X}", checked=bool(prop['get']['cb'](domw)&flags[fl]), height=1)
                    fcb.stateChanged.connect(_boundFlags(
                                prop['set']['cb'], prop['get']['cb'],
                                domw, lambda v: v==ttk.TTkK.Checked, flags[fl]))
                else:
                    fcb = ttk.TTkCheckbox(text=f" 0x{flags[fl]:04X}", checked=bool(prop['get']['cb'](domw)&flags[fl]), height=1, enabled=False)
                ret.addChild(ttk.TTkTreeWidgetItem([f"{fl}", fcb]))
            return ret

        # Single Flag Fields
        # • Check State  │[Partially Checked    ^]│
        #                │┌───────────────────┐   │
        #                ││Checked            │   │
        #                ││Unchecked          │   │
        #                ││Partially Checked  │   │
        #                │└───────────────────┘   │
        def _processSingleFlag(name, prop):
            flags = prop['get']['flags']
            items = [(k,v) for k,v in flags.items()]
            if 'set' in prop:
                value = ttk.TTkComboBox(list=[n for n,_ in items], height=1, textAlign=ttk.TTkK.LEFT_ALIGN)
                value.setCurrentIndex([cs for _,cs in items].index(prop['get']['cb'](domw)))
                value.currentTextChanged.connect(_bound(prop['set']['cb'],domw, lambda v:flags[v]))
            else:
                value = ttk.TTkLabel(text=items[[cs for _,cs in items].index(prop['get']['cb'](domw))][0])
            return ttk.TTkTreeWidgetItem([name,value])

        # List Fields
        # property in this format:
        #    'Position' : {
        #            'init': {'name':'pos', 'type': [
        #                            { 'name': 'x', 'type':int } ,
        #                            { 'name': 'y', 'type':int } ] },
        #            'get':  { 'cb':pos,    'type': [
        #                            { 'name': 'x', 'type':int } ,
        #                            { 'name': 'y', 'type':int } ] },
        #            'set':  { 'cb':move,   'type': [
        #                            { 'name': 'x', 'type':int } ,
        #                            { 'name': 'y', 'type':int } ] } },
        #
        def _processList(name, prop):
            def _getter(_i, _p):
                def _ret(_w):
                    return _p['get']['cb'](_w)[_i]
                return _ret

            def _setter(_i, _p):
                def _ret(_w,_v):
                    __vals = list(_p['get']['cb'](_w))
                    __vals[_i]=_v
                    _p['set']['cb'](_w, *__vals)
                return _ret

            curVal = prop['get']['cb'](domw)
            value = ttk.TTkLabel(text=f"{curVal}")
            ret = ttk.TTkTreeWidgetItem([name,value])
            for _i, _prop in enumerate(prop['get']['type']):
                # Defining a proxy property to set or get a single value
                _newProp = {
                    'get' : { 'cb': _getter(_i, prop), 'type':_prop['type'] },
                    'set' : { 'cb': _setter(_i, prop), 'type':_prop['type'] }
                }
                # ret.addChild(ttk.TTkTreeWidgetItem([_prop['name'],f"{curVal[_i]}"]))
                ret.addChild(_processProp(_prop['name'], _newProp))
            return ret

        # Dict Fields
        def _processDict(name, prop):
            curVal = prop['get']['cb'](domw)
            value = ttk.TTkLabel(text=f"{curVal} - TBD")
            ret = ttk.TTkTreeWidgetItem([name,value])
            return ret

        # Boolean Fields
        def _processBool(name, prop):
            getval = prop['get']['cb'](domw)
            value = ttk.TTkCheckbox(text=f" {p}", checked=getval, height=1)
            value.stateChanged.connect(_bound(prop['set']['cb'],domw, lambda v:v==ttk.TTkK.Checked))
            return ttk.TTkTreeWidgetItem([name,value])
        # Integer Fields
        def _processInt(name, prop):
            getval = prop['get']['cb'](domw)
            value = ttk.TTkSpinBox(value=getval, height=1, maximum=0x10000, minimum=-0x10000)
            value.valueChanged.connect(_bound(prop['set']['cb'],domw,lambda v:v))
            return ttk.TTkTreeWidgetItem([name,value])
        # String Fields
        def _processStr(name, prop):
            getval = prop['get']['cb'](domw)
            value = ttk.TTkLineEdit(text=getval, height=len(getval.split('\n')))
            value.textChanged.connect(_boundLineEdit(prop['set']['cb'],domw,value))
            return ttk.TTkTreeWidgetItem([name,value])
        # String Fields
        def _processTTkString(name, prop, multiLine=True):
            getval = prop['get']['cb'](domw)
            value = ttk.TTkTextPicker(text=getval, height=len(getval.split('\n')), autoSize=True, multiLine=multiLine)
            value.textChanged.connect(_boundTextEdit(prop['set']['cb'],domw,value))
            return ttk.TTkTreeWidgetItem([name,value])
        # Color Fields
        def _processTTkColor(name, prop):
            getval = prop['get']['cb'](domw)
            value = ttk.TTkWidget(layout=ttk.TTkHBoxLayout(), height=1)
            value.layout().addWidget(_cb := ttk.TTkColorButtonPicker(color=getval, height=1))
            value.layout().addWidget(_rc := ttk.TTkButton(text=ttk.TTkString('x',ttk.TTkColor.fg('#FFAA00')),maxWidth=3))
            _cb.colorSelected.connect(_bound(prop['set']['cb'],domw,lambda v:v))
            _rc.clicked.connect(_boundValue(prop['set']['cb'],domw,ttk.TTkColor.RST))
            _rc.clicked.connect(lambda :_cb.setColor(ttk.TTkColor.RST))
            return ttk.TTkTreeWidgetItem([name,value])
        # Layout field
        def _processTTkLayout(name, prop):
            value = ttk.TTkComboBox(list=['TTkLayout','TTkGridLayout','TTkHBoxLayout','TTkVBoxLayout'], height=1)
            value.setCurrentText(prop['get']['cb'](domw).__class__.__name__)
            value.currentTextChanged.connect(_bound(prop['set']['cb'],domw, lambda v:globals()[v]()))
            return ttk.TTkTreeWidgetItem([name,value])

        # Unrecognised Field
        def _processUnknown(name, prop):
            getval = prop['get']['cb'](domw)
            if type(prop['get']['type']) == str:
                getval = f"{prop['get']['type']} = {getval}"
            elif issubclass(prop['get']['type'], ttk.TTkLayout):
                getval = getval.__class__.__name__
            value = ttk.TTkLabel(minSize=(30,1), maxHeight=1, text=f"{getval}", height=1)
            return ttk.TTkTreeWidgetItem([name,value])

        def _processProp(name, prop):
            if 'get' in prop:
                if prop['get']['type'] == 'multiflags':
                    return _processMultiFlag(name, prop)
                elif prop['get']['type'] == 'singleflag':
                    return _processSingleFlag(name, prop)
                elif prop['get']['type'] == bool and 'set' in prop:
                    return _processBool(name, prop)
                elif prop['get']['type'] == int and 'set' in prop:
                    return _processInt(name, prop)
                elif prop['get']['type'] == str and 'set' in prop:
                    return _processStr(name, prop)
                elif prop['get']['type'] == ttk.TTkString and 'set' in prop:
                    return _processTTkString(p,prop,multiLine=True)
                elif prop['get']['type'] == 'singleLineTTkString':
                    return _processTTkString(p,prop,multiLine=False)
                elif prop['get']['type'] == ttk.TTkColor and 'set' in prop:
                    return _processTTkColor(name, prop)
                elif prop['get']['type'] == ttk.TTkLayout and 'set' in prop:
                    return _processTTkLayout(name, prop)
                elif type(prop['get']['type']) == list:
                    return _processList(name, prop)
                elif type(prop['get']['type']) == dict:
                    return _processDict(name, prop)
                else:
                    return _processUnknown(name, prop)

        proplist = exclude
        self._detail.clear()
        for cc in reversed(type(domw).__mro__):
            # if hasattr(cc,'_ttkProperties'):
            if issubclass(cc, ttk.TTkWidget) or issubclass(cc, ttk.TTkLayout):
                ccName = cc.__name__
                classItem = ttk.TTkTreeWidgetItem([ccName,''], expanded=True)
                self._detail.addTopLevelItem(classItem)
                if ccName in ttk.TTkUiProperties:
                    for p in ttk.TTkUiProperties[ccName]['properties']:
                        if p in exceptions:
                            prop = exceptions[p]
                        else:
                            prop = ttk.TTkUiProperties[ccName]['properties'][p]
                        if p not in proplist:
                            proplist.append(p)
                            classItem.addChild(_processProp(p, prop))



# MIT License
#
# Copyright (c) 2023 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import TermTk as ttk

from .cfg  import *
from .about import *

dWidgets = {
    'Layouts':{
        "Layout"       : { "class":ttk.TTkLayout    , "params":{'size':(30,10)}},
        "H Box Layout" : { "class":ttk.TTkHBoxLayout, "params":{'size':(30,10)}},
        "V Box Layout" : { "class":ttk.TTkVBoxLayout, "params":{'size':(30,10)}},
        "Grid Layout"  : { "class":ttk.TTkGridLayout, "params":{'size':(30,10)}},
        "Splitter"     : { "class":ttk.TTkSplitter  , "params":{'size':(40,10)}, "disabled": True},
    },
    'Buttons':{
        "Button"       : { "class":ttk.TTkButton,      "params":{'size':(20,3), 'text':'Button', 'border':True, }},
        "Button Slim"  : { "class":ttk.TTkButton,      "params":{'size':(20,1), 'text':'Button', 'border':False, }},
        "Checkbox"     : { "class":ttk.TTkCheckbox,    "params":{'size':(20,1), 'text':'Checkbox' }},
        "Radio Button" : { "class":ttk.TTkRadioButton, "params":{'size':(20,1), 'text':'Radio b.', 'radiogroup':'DefaultGroup' }},
    },
    'Input Widgets':{
        "ComboBox"    : { "class":ttk.TTkComboBox,  "params":{'size':(20,1)} },
        "LineEdit"    : { "class":ttk.TTkLineEdit,  "params":{'size':(20,1)} },
        "TextEdit"    : { "class":ttk.TTkTextEdit,  "params":{'size':(20,5), 'readOnly':False, 'multiline':True } },
        "TextEditLine": { "class":ttk.TTkTextEdit,  "params":{'size':(20,1), 'readOnly':False, 'multiLine':False, 'maxHeight':1 } },
        "SpinBox"     : { "class":ttk.TTkSpinBox,   "params":{'size':(20,1)} },
        "H ScrollBar" : { "class":ttk.TTkScrollBar, "params":{'size':(10,1), "orientation":ttk.TTkK.HORIZONTAL} },
        "V ScrollBar" : { "class":ttk.TTkScrollBar, "params":{'size':(1,5),  "orientation":ttk.TTkK.VERTICAL} },
    },
    'Widgets':{
        "Label"           : { "class":ttk.TTkLabel,          "params":{'size':(20,1), 'text':'Label'}},
        "List"            : { "class":ttk.TTkListWidget,     "params":{'size':(20,1)}, "disabled": True},
        "Scroll Area"     : { "class":ttk.TTkScrollArea,     "params":{'size':(20,5)}, "disabled": True},
        "Spacer"          : { "class":ttk.TTkSpacer,         "params":{'size':(10,5)}},
        "Tab Widget"      : { "class":ttk.TTkTabWidget,      "params":{'size':(20,3)}, "disabled": True},
        "Window"          : { "class":ttk.TTkWindow,         "params":{'size':(20,10)}},
        "Widget"          : { "class":ttk.TTkWidget,         "params":{'size':(20,5)}},
        "Frame"           : { "class":ttk.TTkFrame,          "params":{'size':(20,5), 'border':True}},
        "Resizable Frame" : { "class":ttk.TTkResizableFrame, "params":{'size':(20,5)}},
    },
    'Pickers':{
        "Color Picker"     : { "class":ttk.TTkColorButtonPicker, "params":{'size':( 6,3), 'border':True}},
        "File Picker"      : { "class":ttk.TTkFileButtonPicker,  "params":{'size':(20,3), 'border':True}},
        "Date Picker"      : { "class":ttk.TTkButton, "params":{'size':(20,3)}, "disabled": True},
        "TtkString Picker" : { "class":ttk.TTkButton, "params":{'size':(20,3)}, "disabled": True},
    },
    'Debug':{
        "Log Viewer"       : { "class":ttk.TTkLogViewer,       "params":{'size':(60,10)}},
        "Input View"       : { "class":ttk.TTkKeyPressView,    "params":{'size':(60,3)}},
        "Tom Inspector"    : { "class":ttk.TTkTomInspector,    "params":{'size':(40,10)}, "disabled": True},
        "Test Widget"      : { "class":ttk.TTkTestWidgets,     "params":{'size':(40,10)}, "disabled": True},
        "Test Widget info" : { "class":ttk.TTkTestWidgetSizes, "params":{'size':(40,10)}},
    }
}

class DragDesignItem(ttk.TTkWidget):
    __slots__ = ('_itemName', '_widgetClass', '_designer')
    def __init__(self, itemName, widgetClass, designer, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setMaximumHeight(3)
        self.setMinimumSize(max(15,len(itemName)+2),3)
        self._itemName = itemName
        self._widgetClass = widgetClass
        self._designer = designer
        self.setEnabled('disabled' not in widgetClass)

    def mouseDragEvent(self, evt) -> bool:
        ttk.TTkLog.debug(f"Start DnD -> {self._itemName}")
        wc = self._widgetClass

        # Gen unique name for the new widget
        basename = wc['class'].__name__
        name = basename
        index = 1
        names = {w.name() for w in self._designer.getWidgets()}
        while name in names:
            name = f"{basename}-{index}"
            index += 1

        drag = ttk.TTkDrag()
        data = wc['class'](**(wc['params']|{'name':name}))
        if issubclass(wc['class'], ttk.TTkWidget):
            drag.setPixmap(data)
        else:
            w,h = wc['params']['size']
            pm = ttk.TTkCanvas(width=w, height=h)
            pm.drawBox(pos=(0,0),size=(w,h), color=ttk.TTkColor.fg('#888888'))
            drag.setPixmap(pm)
        drag.setData(data)
        drag.exec()
        return True

    def paintEvent(self):
        if self.isEnabled():
            color=ttk.TTkColor.RST
        else:
            color=ttk.TTkColor.fg('#AAAAAA')
        self._canvas.drawText(text=self._itemName, pos=(1,1), color=color)
        self._canvas.drawBox(pos=(0,0),size=self.size(), color=color)

class WidgetBox(ttk.TTkVBoxLayout):
    def __init__(self, designer, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for tw in dWidgets:
            self.addWidget(ttk.TTkLabel(text=tw, color=ttk.TTkColor.fg('#FFFF88')+ttk.TTkColor.bg('#0000FF')))
            for ww in dWidgets[tw]:
                self.addWidget(DragDesignItem(ww, dWidgets[tw][ww], designer))
        # self.setGeometry(0,0,self.minimumWidth(), self.minimumHeight())

class WidgetBoxScrollArea(ttk.TTkScrollArea):
    def __init__(self, designer,*args, **kwargs):
        super().__init__(*args, **kwargs)
        self.viewport().setLayout(WidgetBox(designer))
